/* eslint-disable @typescript-eslint/no-shadow */
import {
  clone,
  pick,
  filterChange,
  hasLocalStorage,
  changesHandler as Changes,
  toPromise,
  uuid,
} from 'pouchdb-utils';

import { isDeleted, isLocalId } from 'pouchdb-adapter-utils';
import { collectConflicts, traverseRevTree, latest as getLatest } from 'pouchdb-merge';
import { safeJsonParse, safeJsonStringify } from 'pouchdb-json';
import { MISSING_DOC, REV_CONFLICT, WSQ_ERROR, createError } from 'pouchdb-errors';
import { binaryStringToBlobOrBuffer as binStringToBlob, btoa } from 'pouchdb-binary-utils';

import parseHexString from './parseHex';
import websqlBulkDocs from './bulkDocs';

import {
  ADAPTER_VERSION,
  DOC_STORE,
  BY_SEQ_STORE,
  ATTACH_STORE,
  LOCAL_STORE,
  META_STORE,
  ATTACH_AND_SEQ_STORE,
} from './constants';

import {
  qMarks,
  stringifyDoc,
  unstringifyDoc,
  select,
  compactRevs,
  websqlError,
  getSize,
  unescapeBlob,
} from './utils';

import openDB from './openDatabase';

export type PouchDocument = PouchDB.Core.Document<any> & PouchDB.Core.GetMeta;
export type PouchAllDocs = PouchDB.Core.AllDocsWithKeyOptions &
  PouchDB.Core.AllDocsWithKeysOptions &
  PouchDB.Core.AllDocsWithinRangeOptions &
  PouchDB.Core.AllDocsOptions;

export interface SQLPouchAdapter {
  _id: Promise<string>;
  _db: Database;
  _name: string;
  _remote: boolean;

  auto_compaction: boolean;

  type(): string;

  _close(callback): void;

  _info(callback: (error: any, info?: Partial<PouchDB.Core.DatabaseInfo>) => void): void;

  _getLocal(id: PouchDB.Core.DocumentId, callback): void;

  _putLocal(doc: PouchDocument, callback): void;

  _putLocal(doc: PouchDocument, opts: Partial<{ ctx: SQLTransaction }>, callback): void;

  _removeLocal(doc: PouchDocument, opts: Partial<{ ctx: SQLTransaction }>, callback): void;

  _destroy(opts, callback): void;

  _changes(opts): { cancel(): void };

  _allDocs(opts: PouchAllDocs, callback): void;

  _bulkDocs(req, reqOpts, callback): void;

  _get(docId: PouchDB.Core.DocumentId, assign: any, callback): void;

  _getRevisionTree(docId: PouchDB.Core.DocumentId, callback): void;

  _doCompaction(docId: PouchDB.Core.DocumentId, revs: string[], callback): any;

  _getAttachment(
    _id: PouchDB.Core.DocumentId,
    att: PouchDB.Core.AttachmentId,
    attObj: PouchDB.Core.Attachment,
    attOpts: { ctx: SQLTransaction; binary: boolean },
    cb: PouchDB.Core.Callback<Blob | Buffer>
  ): void;
}

const websqlChanges = new Changes();

function fetchAttachmentsIfNecessary(
  document: PouchDocument,
  opts: PouchAllDocs,
  api: SQLPouchAdapter,
  txn: SQLTransaction,
  cb?
) {
  const attachments = Object.keys(document._attachments || {});

  if (!attachments.length) {
    return cb && cb();
  }

  let numDone = 0;

  function checkDone() {
    if (++numDone === attachments.length && cb) {
      cb();
    }
  }

  function fetchAttachment<T>(doc: PouchDocument, att: string) {
    const attObj = doc._attachments[att];
    const attOpts = { binary: opts.binary, ctx: txn };

    api._getAttachment(doc._id, att, attObj, attOpts, (_, data) => {
      doc._attachments[att] = Object.assign(pick(attObj, ['digest', 'content_type']), { data });
      checkDone();
    });
  }

  attachments.forEach((att) => {
    if (opts.attachments && opts.include_docs) {
      fetchAttachment(document, att);
    } else {
      document._attachments[att].stub = true;
      checkDone();
    }
  });
}

const POUCH_VERSION = 1;

// these indexes cover the ground for most allDocs queries
const BY_SEQ_STORE_DELETED_INDEX_SQL =
  "CREATE INDEX IF NOT EXISTS 'by-seq-deleted-idx' ON " + BY_SEQ_STORE + ' (seq, deleted)';
const BY_SEQ_STORE_DOC_ID_REV_INDEX_SQL =
  "CREATE UNIQUE INDEX IF NOT EXISTS 'by-seq-doc-id-rev' ON " + BY_SEQ_STORE + ' (doc_id, rev)';
const DOC_STORE_WINNINGSEQ_INDEX_SQL =
  "CREATE INDEX IF NOT EXISTS 'doc-winningseq-idx' ON " + DOC_STORE + ' (winningseq)';
const ATTACH_AND_SEQ_STORE_SEQ_INDEX_SQL =
  "CREATE INDEX IF NOT EXISTS 'attach-seq-seq-idx' ON " + ATTACH_AND_SEQ_STORE + ' (seq)';
const ATTACH_AND_SEQ_STORE_ATTACH_INDEX_SQL =
  "CREATE UNIQUE INDEX IF NOT EXISTS 'attach-seq-digest-idx' ON " +
  ATTACH_AND_SEQ_STORE +
  ' (digest, seq)';

const DOC_STORE_AND_BY_SEQ_JOINER = BY_SEQ_STORE + '.seq = ' + DOC_STORE + '.winningseq';

const SELECT_DOCS =
  BY_SEQ_STORE +
  '.seq AS seq, ' +
  BY_SEQ_STORE +
  '.deleted AS deleted, ' +
  BY_SEQ_STORE +
  '.json AS data, ' +
  BY_SEQ_STORE +
  '.rev AS rev, ' +
  DOC_STORE +
  '.json AS metadata';

function WebSqlPouch(
  opts: PouchDB.Configuration.DatabaseConfiguration,
  cb: { (_: null, api: any): void; (error: any): void }
) {
  const api: SQLPouchAdapter = this;
  const size = getSize(opts);
  const idRequests = [];

  let instanceId: string | null = null;
  let encoding;

  api._name = opts.name;

  // extend the options here, because sqlite plugin has a ton of options
  // and they are constantly changing, so it's more prudent to allow anything
  const websqlOpts = Object.assign({}, opts, {
    version: POUCH_VERSION,
    description: opts.name,
    size,
  });

  const openDBResult = openDB(websqlOpts);
  if (openDBResult.error) {
    return websqlError(cb)(openDBResult.error);
  }

  const db = openDBResult.db;
  if (typeof db.readTransaction !== 'function') {
    // doesn't exist in sqlite plugin
    db.readTransaction = db.transaction;
  }

  api._db = db;

  function dbCreated() {
    // note the db name in case the browser upgrades to idb
    if (hasLocalStorage()) {
      window.localStorage['_pouch__websqldb_' + api._name] = true;
    }

    cb(null, api);
  }

  // In this migration, we added the 'deleted' and 'local' columns to the
  // by-seq and doc store tables.
  // To preserve existing user data, we re-process all the existing JSON
  // and add these values.
  // Called migration2 because it corresponds to adapter version (db_version) #2
  function runMigration2(tx: SQLTransaction, callback) {
    // index used for the join in the allDocs query
    tx.executeSql(DOC_STORE_WINNINGSEQ_INDEX_SQL);

    tx.executeSql(
      'ALTER TABLE ' + BY_SEQ_STORE + ' ADD COLUMN deleted TINYINT(1) DEFAULT 0',
      [],
      function () {
        tx.executeSql(BY_SEQ_STORE_DELETED_INDEX_SQL);

        tx.executeSql(
          'ALTER TABLE ' + DOC_STORE + ' ADD COLUMN local TINYINT(1) DEFAULT 0',
          [],
          function () {
            tx.executeSql(
              "CREATE INDEX IF NOT EXISTS 'doc-store-local-idx' ON " + DOC_STORE + ' (local, id)'
            );

            const sql =
              'SELECT ' +
              DOC_STORE +
              '.winningseq AS seq, ' +
              DOC_STORE +
              '.json AS metadata FROM ' +
              BY_SEQ_STORE +
              ' JOIN ' +
              DOC_STORE +
              ' ON ' +
              BY_SEQ_STORE +
              '.seq = ' +
              DOC_STORE +
              '.winningseq';

            tx.executeSql(sql, [], function (tx, result) {
              const deleted = [];
              const local = [];

              for (let i = 0; i < result.rows.length; i++) {
                const item = result.rows.item(i);
                const seq = item.seq;
                const metadata = JSON.parse(item.metadata);

                if (isDeleted(metadata)) {
                  deleted.push(seq);
                }

                if (isLocalId(metadata.id)) {
                  local.push(metadata.id);
                }
              }

              tx.executeSql(
                'UPDATE ' + DOC_STORE + 'SET local = 1 WHERE id IN ' + qMarks(local.length),
                local,
                function () {
                  tx.executeSql(
                    'UPDATE ' +
                      BY_SEQ_STORE +
                      ' SET deleted = 1 WHERE seq IN ' +
                      qMarks(deleted.length),
                    deleted,
                    callback
                  );
                }
              );
            });
          }
        );
      }
    );
  }

  // in this migration, we make all the local docs unversioned
  function runMigration3(tx: SQLTransaction, callback) {
    const local = 'CREATE TABLE IF NOT EXISTS ' + LOCAL_STORE + ' (id UNIQUE, rev, json)';

    tx.executeSql(local, [], function () {
      const sql =
        'SELECT ' +
        DOC_STORE +
        '.id AS id, ' +
        BY_SEQ_STORE +
        '.json AS data ' +
        'FROM ' +
        BY_SEQ_STORE +
        ' JOIN ' +
        DOC_STORE +
        ' ON ' +
        BY_SEQ_STORE +
        '.seq = ' +
        DOC_STORE +
        '.winningseq WHERE local = 1';

      tx.executeSql(sql, [], function (tx, res) {
        const rows = [];

        for (let i = 0; i < res.rows.length; i++) {
          rows.push(res.rows.item(i));
        }

        function doNext() {
          if (!rows.length) {
            return callback(tx);
          }

          const row = rows.shift();
          const rev = JSON.parse(row.data)._rev;

          tx.executeSql(
            'INSERT INTO ' + LOCAL_STORE + ' (id, rev, json) VALUES (?,?,?)',
            [row.id, rev, row.data],
            function (tx: SQLTransaction) {
              tx.executeSql(
                'DELETE FROM ' + DOC_STORE + ' WHERE id=?',
                [row.id],
                function (tx: SQLTransaction) {
                  tx.executeSql(
                    'DELETE FROM ' + BY_SEQ_STORE + ' WHERE seq=?',
                    [row.seq],
                    function () {
                      doNext();
                    }
                  );
                }
              );
            }
          );
        }

        doNext();
      });
    });
  }

  // in this migration, we remove doc_id_rev and just use rev
  function runMigration4(tx: SQLTransaction, callback) {
    function updateRows(rows) {
      function doNext() {
        if (!rows.length) {
          return callback(tx);
        }

        const row = rows.shift();
        const doc_id_rev = parseHexString(row.hex, encoding);
        const idx = doc_id_rev.lastIndexOf('::');
        const doc_id = doc_id_rev.substring(0, idx);
        const rev = doc_id_rev.substring(idx + 2);
        const sql = 'UPDATE ' + BY_SEQ_STORE + ' SET doc_id=?, rev=? WHERE doc_id_rev=?';

        tx.executeSql(sql, [doc_id, rev, doc_id_rev], function () {
          doNext();
        });
      }

      doNext();
    }

    const sql = 'ALTER TABLE ' + BY_SEQ_STORE + ' ADD COLUMN doc_id';
    tx.executeSql(sql, [], function (tx) {
      const sql = 'ALTER TABLE ' + BY_SEQ_STORE + ' ADD COLUMN rev';

      tx.executeSql(sql, [], function (tx) {
        tx.executeSql(BY_SEQ_STORE_DOC_ID_REV_INDEX_SQL, [], function (tx) {
          const sql = 'SELECT hex(doc_id_rev) as hex FROM ' + BY_SEQ_STORE;

          tx.executeSql(sql, [], function (tx, res) {
            const rows = [];

            for (let i = 0; i < res.rows.length; i++) {
              rows.push(res.rows.item(i));
            }

            updateRows(rows);
          });
        });
      });
    });
  }

  // in this migration, we add the attach_and_seq table
  // for issue #2818
  function runMigration5(tx: SQLTransaction, callback) {
    function migrateAttsAndSeqs(tx: SQLTransaction) {
      // need to actually populate the table. this is the expensive part,
      // so as an optimization, check first that this database even
      // contains attachments
      const sql = 'SELECT COUNT(*) AS cnt FROM ' + ATTACH_STORE;

      tx.executeSql(sql, [], function (tx, res) {
        const count = res.rows.item(0).cnt;

        if (!count) {
          return callback(tx);
        }

        let offset = 0;
        const pageSize = 10;

        function nextPage() {
          let sql = select(
            SELECT_DOCS + ', ' + DOC_STORE + '.id AS id',
            [DOC_STORE, BY_SEQ_STORE],
            DOC_STORE_AND_BY_SEQ_JOINER,
            null,
            DOC_STORE + '.id '
          );

          sql += ' LIMIT ' + pageSize + ' OFFSET ' + offset;
          offset += pageSize;

          tx.executeSql(sql, [], function (tx, res) {
            if (!res.rows.length) {
              return callback(tx);
            }

            const digestSeqs = {};

            function addDigestSeq(digest, seq) {
              // uniq digest/seq pairs, just in case there are dups
              const seqs = (digestSeqs[digest] = digestSeqs[digest] || []);
              if (seqs.indexOf(seq) === -1) {
                seqs.push(seq);
              }
            }

            for (let i = 0; i < res.rows.length; i++) {
              const row = res.rows.item(i);
              const doc = unstringifyDoc(row.data, row.id, row.rev);
              const atts = Object.keys(doc._attachments || {});

              for (const item of atts) {
                const att = doc._attachments[item];
                addDigestSeq(att.digest, row.seq);
              }
            }

            const digestSeqPairs = [];

            Object.keys(digestSeqs).forEach(function (digest) {
              const seqs = digestSeqs[digest];

              seqs.forEach(function (seq) {
                digestSeqPairs.push([digest, seq]);
              });
            });

            if (!digestSeqPairs.length) {
              return nextPage();
            }

            let numDone = 0;
            digestSeqPairs.forEach(function (pair) {
              const sql = 'INSERT INTO ' + ATTACH_AND_SEQ_STORE + ' (digest, seq) VALUES (?,?)';

              tx.executeSql(sql, pair, function () {
                if (++numDone === digestSeqPairs.length) {
                  nextPage();
                }
              });
            });
          });
        }

        nextPage();
      });
    }

    const attachAndRev =
      'CREATE TABLE IF NOT EXISTS ' + ATTACH_AND_SEQ_STORE + ' (digest, seq INTEGER)';

    tx.executeSql(attachAndRev, [], function (tx) {
      tx.executeSql(ATTACH_AND_SEQ_STORE_ATTACH_INDEX_SQL, [], function (tx) {
        tx.executeSql(ATTACH_AND_SEQ_STORE_SEQ_INDEX_SQL, [], migrateAttsAndSeqs);
      });
    });
  }

  // in this migration, we use escapeBlob() and unescapeBlob()
  // instead of reading out the binary as HEX, which is slow
  function runMigration6(tx: SQLTransaction, callback) {
    const sql = 'ALTER TABLE ' + ATTACH_STORE + ' ADD COLUMN escaped TINYINT(1) DEFAULT 0';
    tx.executeSql(sql, [], callback);
  }

  // issue #3136, in this migration we need a "latest seq" as well
  // as the "winning seq" in the doc store
  function runMigration7(tx: SQLTransaction, callback) {
    const sql = 'ALTER TABLE ' + DOC_STORE + ' ADD COLUMN max_seq INTEGER';
    tx.executeSql(sql, [], function (tx) {
      const sql =
        'UPDATE ' +
        DOC_STORE +
        ' SET max_seq=(SELECT MAX(seq) FROM ' +
        BY_SEQ_STORE +
        ' WHERE doc_id=id)';
      tx.executeSql(sql, [], function (tx) {
        // add unique index after filling, else we'll get a constraint
        // error when we do the ALTER TABLE
        const sql =
          "CREATE UNIQUE INDEX IF NOT EXISTS 'doc-max-seq-idx' ON " + DOC_STORE + ' (max_seq)';
        tx.executeSql(sql, [], callback);
      });
    });
  }

  function checkEncoding(tx: SQLTransaction, callback) {
    // UTF-8 on chrome/android, UTF-16 on safari < 7.1
    tx.executeSql('SELECT HEX("a") AS hex', [], function (tx, res) {
      const hex = res.rows.item(0).hex;
      encoding = hex.length === 2 ? 'UTF-8' : 'UTF-16';
      callback();
    });
  }

  function onGetInstanceId() {
    while (idRequests.length > 0) {
      const idCallback = idRequests.pop();
      idCallback(null, instanceId);
    }
  }

  function onGetVersion(tx: SQLTransaction, dbVersion: number) {
    if (dbVersion === 0) {
      // initial schema
      const meta = 'CREATE TABLE IF NOT EXISTS ' + META_STORE + ' (dbid, db_version INTEGER)';
      const attach =
        'CREATE TABLE IF NOT EXISTS ' +
        ATTACH_STORE +
        ' (digest UNIQUE, escaped TINYINT(1), body BLOB)';

      const attachAndRev =
        'CREATE TABLE IF NOT EXISTS ' + ATTACH_AND_SEQ_STORE + ' (digest, seq INTEGER)';

      // TODO: migrate winningseq to INTEGER
      const doc =
        'CREATE TABLE IF NOT EXISTS ' +
        DOC_STORE +
        ' (id unique, json, winningseq, max_seq INTEGER UNIQUE)';

      const seq =
        'CREATE TABLE IF NOT EXISTS ' +
        BY_SEQ_STORE +
        ' (seq INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, ' +
        'json, deleted TINYINT(1), doc_id, rev)';

      const local = 'CREATE TABLE IF NOT EXISTS ' + LOCAL_STORE + ' (id UNIQUE, rev, json)';

      // creates
      tx.executeSql(attach);
      tx.executeSql(local);

      tx.executeSql(attachAndRev, [], function () {
        tx.executeSql(ATTACH_AND_SEQ_STORE_SEQ_INDEX_SQL);
        tx.executeSql(ATTACH_AND_SEQ_STORE_ATTACH_INDEX_SQL);
      });

      tx.executeSql(doc, [], function () {
        tx.executeSql(DOC_STORE_WINNINGSEQ_INDEX_SQL);

        tx.executeSql(seq, [], function () {
          tx.executeSql(BY_SEQ_STORE_DELETED_INDEX_SQL);
          tx.executeSql(BY_SEQ_STORE_DOC_ID_REV_INDEX_SQL);

          tx.executeSql(meta, [], function () {
            instanceId = uuid();

            // mark the db version, and new dbid
            const initSeq = 'INSERT INTO ' + META_STORE + ' (db_version, dbid) VALUES (?,?)';
            const initSeqArgs = [ADAPTER_VERSION, instanceId];

            tx.executeSql(initSeq, initSeqArgs, function () {
              onGetInstanceId();
            });
          });
        });
      });
    } else {
      // version > 0

      const setupDone = function () {
        const migrated = dbVersion < ADAPTER_VERSION;

        if (migrated) {
          // update the db version within this transaction
          tx.executeSql('UPDATE ' + META_STORE + ' SET db_version = ' + ADAPTER_VERSION);
        }

        // notify db.id() callers
        const sql = 'SELECT dbid FROM ' + META_STORE;

        tx.executeSql(sql, [], function (tx, result) {
          instanceId = result.rows.item(0).dbid;
          onGetInstanceId();
        });
      };

      // would love to use promises here, but then websql
      // ends the transaction early
      const tasks = [
        runMigration2,
        runMigration3,
        runMigration4,
        runMigration5,
        runMigration6,
        runMigration7,
        setupDone,
      ];

      // run each migration sequentially
      let i = dbVersion;

      const nextMigration = function (tx: SQLTransaction) {
        tasks[i - 1](tx, nextMigration);
        i++;
      };

      nextMigration(tx);
    }
  }

  function setup() {
    db.transaction(
      function (tx) {
        // first check the encoding
        checkEncoding(tx, function () {
          // then get the version
          fetchVersion(tx);
        });
      },
      websqlError(cb),
      dbCreated
    );
  }

  function fetchVersion(tx: SQLTransaction) {
    const sql = 'SELECT sql FROM sqlite_master WHERE tbl_name = ' + META_STORE;

    tx.executeSql(sql, [], function (tx, result) {
      if (!result.rows.length) {
        // database hasn't even been created yet (version 0)
        onGetVersion(tx, 0);
      } else if (!/db_version/.test(result.rows.item(0).sql)) {
        // table was created, but without the new db_version column,
        // so add it.
        tx.executeSql(
          'ALTER TABLE ' + META_STORE + ' ADD COLUMN db_version INTEGER',
          [],
          function () {
            // before version 2, this column didn't even exist
            onGetVersion(tx, 1);
          }
        );
      } else {
        // column exists, we can safely get it
        tx.executeSql('SELECT db_version FROM ' + META_STORE, [], function (tx, result) {
          const dbVersion = result.rows.item(0).db_version;
          onGetVersion(tx, dbVersion);
        });
      }
    });
  }

  setup();

  function getMaxSeq(tx: SQLTransaction, callback) {
    const sql = 'SELECT MAX(seq) AS seq FROM ' + BY_SEQ_STORE;

    tx.executeSql(sql, [], function (tx, res) {
      const updateSeq = res.rows.item(0).seq || 0;
      callback(updateSeq);
    });
  }

  function countDocs(tx: SQLTransaction, callback) {
    // count the total rows
    const sql = select(
      'COUNT(' + DOC_STORE + ".id) AS 'num'",
      [DOC_STORE, BY_SEQ_STORE],
      DOC_STORE_AND_BY_SEQ_JOINER,
      BY_SEQ_STORE + '.deleted=0'
    );

    tx.executeSql(sql, [], function (tx, result) {
      callback(result.rows.item(0).num);
    });
  }

  api._remote = false;
  api.type = function () {
    return 'websql';
  };

  api._id = toPromise(function (callback) {
    callback(null, instanceId);
  });

  api._info = function (callback) {
    let seq;
    let docCount;

    db.readTransaction(
      function (tx) {
        getMaxSeq(tx, function (theSeq) {
          seq = theSeq;
        });

        countDocs(tx, function (theDocCount) {
          docCount = theDocCount;
        });
      },
      websqlError(callback),
      function () {
        callback(null, {
          doc_count: docCount,
          update_seq: seq,
          websql_encoding: encoding,
        });
      }
    );
  };

  api._bulkDocs = function (req, reqOpts, callback) {
    websqlBulkDocs(opts, req, reqOpts, api, db, websqlChanges, callback);
  };

  function latest(
    tx: SQLTransaction,
    id: PouchDB.Core.DocumentId,
    rev: PouchDB.Core.RevisionId,
    callback,
    finish
  ) {
    const sql = select(
      SELECT_DOCS,
      [DOC_STORE, BY_SEQ_STORE],
      DOC_STORE_AND_BY_SEQ_JOINER,
      DOC_STORE + '.id=?'
    );
    const sqlArgs = [id];

    tx.executeSql(sql, sqlArgs, function (a, results) {
      if (!results.rows.length) {
        const err = createError(MISSING_DOC, 'missing');
        return finish(err);
      }

      const item = results.rows.item(0);
      const metadata = safeJsonParse(item.metadata);

      callback(getLatest(rev, metadata));
    });
  }

  api._get = function (
    id: PouchDB.Core.DocumentId,
    opts: PouchDB.Core.GetOptions & { ctx: SQLTransaction },
    callback
  ) {
    let doc;
    let metadata;
    const tx = opts.ctx;

    if (!tx) {
      return db.readTransaction(function (txn) {
        api._get(id, Object.assign({ ctx: txn }, opts), callback);
      });
    }

    function finish(err?) {
      callback(err, { doc, metadata, ctx: tx });
    }

    let sql;
    let sqlArgs;

    if (!opts.rev) {
      sql = select(
        SELECT_DOCS,
        [DOC_STORE, BY_SEQ_STORE],
        DOC_STORE_AND_BY_SEQ_JOINER,
        DOC_STORE + '.id=?'
      );

      sqlArgs = [id];
    } else if (opts.latest) {
      latest(
        tx,
        id,
        opts.rev,
        function (latestRev) {
          opts.latest = false;
          opts.rev = latestRev;
          api._get(id, opts, callback);
        },
        finish
      );
      return;
    } else {
      sql = select(
        SELECT_DOCS,
        [DOC_STORE, BY_SEQ_STORE],
        DOC_STORE + '.id=' + BY_SEQ_STORE + '.doc_id',
        [BY_SEQ_STORE + '.doc_id=?', BY_SEQ_STORE + '.rev=?']
      );
      sqlArgs = [id, opts.rev];
    }

    tx.executeSql(sql, sqlArgs, function (a, results) {
      if (!results.rows.length) {
        const missingErr = createError(MISSING_DOC, 'missing');
        return finish(missingErr);
      }

      const item = results.rows.item(0);
      metadata = safeJsonParse(item.metadata);

      if (item.deleted && !opts.rev) {
        const deletedErr = createError(MISSING_DOC, 'deleted');
        return finish(deletedErr);
      }

      doc = unstringifyDoc(item.data, metadata.id, item.rev);
      finish();
    });
  };

  api._allDocs = function (opts: PouchAllDocs, callback) {
    const results = [];
    let totalRows;
    let updateSeq;

    const start = 'startkey' in opts ? opts.startkey : false;
    const end = 'endkey' in opts ? opts.endkey : false;
    const key = 'key' in opts ? opts.key : false;
    const keys: Array<string> | false = 'keys' in opts ? opts.keys : false;
    const descending = 'descending' in opts ? opts.descending : false;
    const limit = 'limit' in opts ? opts.limit : -1;
    const offset = 'skip' in opts ? opts.skip : 0;
    const inclusiveEnd = opts.inclusive_end !== false;

    let sqlArgs = [];
    let criteria = [];
    const keyChunks = [];

    if (keys) {
      const distinctKeys = [];

      keys.forEach(function (key) {
        if (distinctKeys.indexOf(key) === -1) {
          distinctKeys.push(key);
        }
      });

      for (let index = 0; index < distinctKeys.length; index += 999) {
        const chunk = distinctKeys.slice(index, index + 999);

        if (chunk.length > 0) {
          keyChunks.push(chunk);
        }
      }
    } else if (key !== false) {
      criteria.push(DOC_STORE + '.id = ?');
      sqlArgs.push(key);
    } else if (start !== false || end !== false) {
      if (start !== false) {
        criteria.push(DOC_STORE + '.id ' + (descending ? '<=' : '>=') + ' ?');
        sqlArgs.push(start);
      }

      if (end !== false) {
        let comparator = descending ? '>' : '<';

        if (inclusiveEnd) {
          comparator += '=';
        }

        criteria.push(DOC_STORE + '.id ' + comparator + ' ?');
        sqlArgs.push(end);
      }

      if (key !== false) {
        criteria.push(DOC_STORE + '.id = ?');
        sqlArgs.push(key);
      }
    }

    if (!keys) {
      // report deleted if keys are specified
      criteria.push(BY_SEQ_STORE + '.deleted = 0');
    }

    db.readTransaction(
      function (tx) {
        // count the docs in parallel to other operations
        countDocs(tx, function (docCount) {
          totalRows = docCount;
        });

        /* istanbul ignore if */
        if (opts.update_seq) {
          // get max sequence in parallel to other operations
          getMaxSeq(tx, function (theSeq) {
            updateSeq = theSeq;
          });
        }

        if (limit === 0) {
          return;
        }

        if (keys) {
          let finishedCount = 0;
          const allRows = [];

          keyChunks.forEach(function (keyChunk) {
            sqlArgs = [];
            criteria = [];
            let bindingStr = '';

            keyChunk.forEach(function () {
              bindingStr += '?,';
            });

            bindingStr = bindingStr.substring(0, bindingStr.length - 1); // keys is never empty
            criteria.push(DOC_STORE + '.id IN (' + bindingStr + ')');
            sqlArgs = sqlArgs.concat(keyChunk);

            let sql = select(
              SELECT_DOCS,
              [DOC_STORE, BY_SEQ_STORE],
              DOC_STORE_AND_BY_SEQ_JOINER,
              criteria,
              DOC_STORE + '.id ' + (descending ? 'DESC' : 'ASC')
            );

            sql += ' LIMIT ' + limit + ' OFFSET ' + offset;

            tx.executeSql(sql, sqlArgs, function (tx, result) {
              finishedCount++;

              for (let index = 0; index < result.rows.length; index++) {
                allRows.push(result.rows.item(index));
              }

              if (finishedCount === keyChunks.length) {
                processResult(allRows);
              }
            });
          });
        } else {
          // do a single query to fetch the documents
          let sql = select(
            SELECT_DOCS,
            [DOC_STORE, BY_SEQ_STORE],
            DOC_STORE_AND_BY_SEQ_JOINER,
            criteria,
            DOC_STORE + '.id ' + (descending ? 'DESC' : 'ASC')
          );

          sql += ' LIMIT ' + limit + ' OFFSET ' + offset;

          tx.executeSql(sql, sqlArgs, function (tx, result) {
            const rows = [];

            for (let index = 0; index < result.rows.length; index++) {
              rows.push(result.rows.item(index));
            }

            processResult(rows);
          });
        }

        function processResult(
          rows: Array<{
            rev: string;
            data: string;
            metadata: PouchDB.Core.AllDocsMeta & { id: string };
            deleted?: boolean | undefined;
          }>
        ) {
          for (let i = 0, l = rows.length; i < l; i++) {
            const item = rows[i];
            const metadata = safeJsonParse(item.metadata);
            const id = metadata.id;
            const data = unstringifyDoc(item.data, id, item.rev);
            const winningRev = data._rev;

            const doc: any = {
              id,
              key: id,
              value: { rev: winningRev },
            };

            if (opts.include_docs) {
              doc.doc = data;
              doc.doc._rev = winningRev;

              if (opts.conflicts) {
                const conflicts = collectConflicts(metadata);

                if (conflicts.length) {
                  doc.doc._conflicts = conflicts;
                }
              }

              fetchAttachmentsIfNecessary(doc.doc, opts, api, tx);
            }

            if (item.deleted) {
              if (keys) {
                doc.value.deleted = true;
                doc.doc = null;
              } else {
                // probably should not happen
                continue;
              }
            }

            if (!keys) {
              results.push(doc);
            } else {
              let index = keys.indexOf(id); // FIXME: indexOf(id, index);

              do {
                results[index] = doc;
                index = keys.indexOf(id, index + 1);
              } while (index > -1 && index < keys.length);
            }
          }

          if (keys) {
            keys.forEach(function (key, index) {
              if (!results[index]) {
                results[index] = { key, error: 'not_found' };
              }
            });
          }
        }
      },
      websqlError(callback),
      function () {
        const returnVal: any = {
          total_rows: totalRows,
          offset: opts.skip,
          rows: results,
        };

        /* istanbul ignore if */
        if (opts.update_seq) {
          returnVal.update_seq = updateSeq;
        }

        callback(null, returnVal);
      }
    );
  };

  api._changes = function (opts: any) {
    opts = clone(opts);

    if (opts.continuous) {
      const id = api._name + ':' + uuid();
      websqlChanges.addListener(api._name, id, api, opts);
      websqlChanges.notify(api._name);

      return {
        cancel() {
          websqlChanges.removeListener(api._name, id);
        },
      };
    }

    const descending = opts.descending;

    // Ignore the `since` parameter when `descending` is true
    opts.since = opts.since && !descending ? opts.since : 0;

    let limit = 'limit' in opts ? opts.limit : -1;
    if (limit === 0) {
      limit = 1; // per CouchDB _changes spec
    }

    const results = [];
    let numResults = 0;

    function fetchChanges() {
      const selectStmt =
        DOC_STORE +
        '.json AS metadata, ' +
        DOC_STORE +
        '.max_seq AS maxSeq, ' +
        BY_SEQ_STORE +
        '.json AS winningDoc, ' +
        BY_SEQ_STORE +
        '.rev AS winningRev ';

      const from = DOC_STORE + ' JOIN ' + BY_SEQ_STORE;

      const joiner =
        DOC_STORE +
        '.id=' +
        BY_SEQ_STORE +
        '.doc_id' +
        ' AND ' +
        DOC_STORE +
        '.winningseq=' +
        BY_SEQ_STORE +
        '.seq';

      const criteria = ['maxSeq > ?'];
      let sqlArgs = [opts.since];

      if (opts.doc_ids) {
        criteria.push(DOC_STORE + '.id IN ' + qMarks(opts.doc_ids.length));
        sqlArgs = sqlArgs.concat(opts.doc_ids);
      }

      const orderBy = 'maxSeq ' + (descending ? 'DESC' : 'ASC');

      let sql = select(selectStmt, from, joiner, criteria, orderBy);

      const filter = filterChange(opts);
      if (!opts.view && !opts.filter) {
        // we can just limit in the query
        sql += ' LIMIT ' + limit;
      }

      let lastSeq = opts.since || 0;
      db.readTransaction(
        function (tx) {
          tx.executeSql(sql, sqlArgs, function (tx, result) {
            function reportChange(change) {
              return function () {
                opts.onChange(change);
              };
            }

            for (let i = 0, l = result.rows.length; i < l; i++) {
              const item = result.rows.item(i);
              const metadata = safeJsonParse(item.metadata);
              lastSeq = item.maxSeq;

              const doc = unstringifyDoc(item.winningDoc, metadata.id, item.winningRev);
              const change = opts.processChange(doc, metadata, opts);
              change.seq = item.maxSeq;

              const filtered = filter(change);
              if (typeof filtered === 'object') {
                return opts.complete(filtered);
              }

              if (filtered) {
                numResults++;

                if (opts.return_docs) {
                  results.push(change);
                }

                // process the attachment immediately
                // for the benefit of live listeners
                if (opts.attachments && opts.include_docs) {
                  fetchAttachmentsIfNecessary(doc, opts, api, tx, reportChange(change));
                } else {
                  reportChange(change)();
                }
              }

              if (numResults === limit) {
                break;
              }
            }
          });
        },
        websqlError(opts.complete),
        function () {
          if (!opts.continuous) {
            opts.complete(null, {
              results,
              last_seq: lastSeq,
            });
          }
        }
      );
    }

    fetchChanges();
  };

  api._close = function (callback) {
    // @ts-ignore
    if (db.close !== undefined) {
      console.log('closing sql database');

      // @ts-ignore
      db.close(
        () => callback(),
        () => callback(createError(WSQ_ERROR, 'failed_to_close'))
      );
    } else {
      // WebSQL databases do not need to be closed
      callback();
    }
  };

  api._getAttachment = function (docId, attachId, attachment, opts, callback) {
    let res;
    const tx = opts.ctx;
    const digest = attachment.digest;
    const type = attachment.content_type;

    const sql =
      'SELECT escaped, ' +
      'CASE WHEN escaped = 1 THEN body ELSE HEX(body) END AS body FROM ' +
      ATTACH_STORE +
      ' WHERE digest=?';

    tx.executeSql(sql, [digest], function (tx, result) {
      // websql has a bug where \u0000 causes early truncation in strings
      // and blobs. to work around this, we used to use the hex() function,
      // but that's not performant. after migration 6, we remove \u0000
      // and add it back in afterwards
      const item = result.rows.item(0);
      const data = item.escaped ? unescapeBlob(item.body) : parseHexString(item.body, encoding);

      if (opts.binary) {
        res = binStringToBlob(data, type);
      } else {
        res = btoa(data);
      }

      callback(null, res);
    });
  };

  api._getRevisionTree = function (docId, callback) {
    db.readTransaction(function (tx) {
      const sql = 'SELECT json AS metadata FROM ' + DOC_STORE + ' WHERE id = ?';

      tx.executeSql(sql, [docId], function (tx, result) {
        if (!result.rows.length) {
          callback(createError(MISSING_DOC));
        } else {
          const data = safeJsonParse(result.rows.item(0).metadata);
          callback(null, data.rev_tree);
        }
      });
    });
  };

  api._doCompaction = function (docId, revs, callback) {
    if (!revs.length) {
      return callback();
    }

    db.transaction(
      function (tx) {
        // update doc store
        const sql = 'SELECT json AS metadata FROM ' + DOC_STORE + ' WHERE id = ?';

        tx.executeSql(sql, [docId], function (tx, result) {
          const metadata = safeJsonParse(result.rows.item(0).metadata);

          traverseRevTree(metadata.rev_tree, function (isLeaf, pos, revHash, ctx, opts) {
            const rev = pos + '-' + revHash;

            if (revs.indexOf(rev) !== -1) {
              opts.status = 'missing';
            }
          });

          const sql = 'UPDATE ' + DOC_STORE + ' SET json = ? WHERE id = ?';
          tx.executeSql(sql, [safeJsonStringify(metadata), docId]);
        });

        compactRevs(revs, docId, tx);
      },
      websqlError(callback),
      function () {
        callback();
      }
    );
  };

  api._getLocal = function (id, callback) {
    db.readTransaction(function (tx) {
      const sql = 'SELECT json, rev FROM ' + LOCAL_STORE + ' WHERE id=?';

      tx.executeSql(sql, [id], function (tx, res) {
        if (res.rows.length) {
          const item = res.rows.item(0);
          const doc = unstringifyDoc(item.json, id, item.rev);
          callback(null, doc);
        } else {
          callback(createError(MISSING_DOC));
        }
      });
    });
  };

  api._putLocal = function (doc, opts, callback?) {
    if (typeof opts === 'function') {
      callback = opts;
      opts = {};
    }

    delete doc._revisions; // ignore this, trust the rev
    const oldRev = doc._rev;
    const id = doc._id;
    let newRev;

    if (!oldRev) {
      newRev = doc._rev = '0-1';
    } else {
      newRev = doc._rev = '0-' + (parseInt(oldRev.split('-')[1], 10) + 1);
    }

    const json = stringifyDoc(doc);
    let ret;

    function putLocal(tx: SQLTransaction) {
      let sql;
      let values;

      if (oldRev) {
        sql = 'UPDATE ' + LOCAL_STORE + ' SET rev=?, json=? ' + 'WHERE id=? AND rev=?';
        values = [newRev, json, id, oldRev];
      } else {
        sql = 'INSERT INTO ' + LOCAL_STORE + ' (id, rev, json) VALUES (?,?,?)';
        values = [id, newRev, json];
      }

      tx.executeSql(
        sql,
        values,
        function (tx, res) {
          if (res.rowsAffected) {
            ret = { ok: true, id, rev: newRev };

            if (opts.ctx) {
              // return immediately
              callback(null, ret);
            }
          } else {
            callback(createError(REV_CONFLICT));
          }
        },
        function () {
          callback(createError(REV_CONFLICT));
          return false; // ack that we handled the error
        }
      );
    }

    if (opts.ctx) {
      putLocal(opts.ctx);
    } else {
      db.transaction(putLocal, websqlError(callback), function () {
        if (ret) {
          callback(null, ret);
        }
      });
    }
  };

  api._removeLocal = function (doc, opts, callback) {
    if (typeof opts === 'function') {
      callback = opts;
      opts = {};
    }

    let ret;

    function removeLocal(tx: SQLTransaction) {
      const sql = 'DELETE FROM ' + LOCAL_STORE + ' WHERE id=? AND rev=?';
      const params = [doc._id, doc._rev];

      tx.executeSql(sql, params, function (tx, res) {
        if (!res.rowsAffected) {
          return callback(createError(MISSING_DOC));
        }

        ret = { ok: true, id: doc._id, rev: '0-0' };
        if (opts.ctx) {
          // return immediately
          callback(null, ret);
        }
      });
    }

    if (opts.ctx) {
      removeLocal(opts.ctx);
    } else {
      db.transaction(removeLocal, websqlError(callback), function () {
        if (ret) {
          callback(null, ret);
        }
      });
    }
  };

  api._destroy = function (opts, callback) {
    websqlChanges.removeAllListeners(api._name);

    db.transaction(
      function (tx) {
        const stores = [
          DOC_STORE,
          BY_SEQ_STORE,
          ATTACH_STORE,
          META_STORE,
          LOCAL_STORE,
          ATTACH_AND_SEQ_STORE,
        ];

        stores.forEach(function (store) {
          tx.executeSql('DROP TABLE IF EXISTS ' + store, []);
        });
      },
      websqlError(callback),
      function () {
        if (hasLocalStorage()) {
          delete window.localStorage['_pouch__websqldb_' + api._name];
          delete window.localStorage[api._name];
        }

        callback(null, { ok: true });
      }
    );
  };
}

export default WebSqlPouch;
